package com.faraphel.tasks_valider.database.api.client.entities.base;

import android.util.Log
import com.faraphel.tasks_valider.database.api.client.TaskEntityHttpClient
import com.faraphel.tasks_valider.database.entities.base.BaseEntity
import com.faraphel.tasks_valider.database.entities.error.HttpException
import com.faraphel.tasks_valider.utils.parser
import com.google.gson.reflect.TypeToken
import kotlin.reflect.KClass
import kotlin.reflect.full.companionObject
import kotlin.reflect.full.companionObjectInstance
import kotlin.reflect.full.declaredMemberProperties


abstract class BaseClientApi<Entity: BaseEntity>(
    private val client: TaskEntityHttpClient,
    private val entityType: KClass<Entity>,
) {
    /**
     * return the API endpoint for this entity
     * @return the API endpoint for this entity
     */
    private fun getEndpoint(): String {
        // get the property for the name of the table
        val propertyTableName = entityType.companionObject!!.declaredMemberProperties.first { member ->
            member.name == "TABLE_NAME"
        }
        // get the table name by calling the getter of the property
        val tableName = propertyTableName.getter.call(entityType.companionObjectInstance)
        // return the endpoint
        return "entities/${tableName}"
    }

    /**
     * return all the entities for that table
     * @return all the entities for that table
     * @throws java.io.IOException reading error while parsing request
     * @throws HttpException error of the request
     */
    fun getAll(): List<Entity> {
        // try to obtain the list of validations
        val response = client.get(this.getEndpoint())

        // in case of error, notify it
        if (!response.isSuccessful)
            throw HttpException(response.code)

        val data = response.body.string()

        // parse the list of validations
        return parser.fromJson(
            data,
            TypeToken.getParameterized(ArrayList::class.java, entityType.java).type
        )
    }

    /**
     * create a new entity in the table
     * @return the id of the object in the database
     * @throws java.io.IOException reading error while parsing request
     * @throws HttpException error of the request
     */
    fun save(entity: Entity): Long {
        // try to send the serialized entity as json
        val response = client.post(
            this.getEndpoint(),
            parser.toJson(entity),
            "application/json; charset=utf-8"
        )

        // in case of error, notify it
        if (!response.isSuccessful)
            throw HttpException(response.code)

        // return the id of the object
        return response.body.string().toLong()
    }

    /**
     * delete an entity in the table
     * @return the number of object deleted in the database
     * @throws java.io.IOException reading error while parsing request
     * @throws HttpException error of the request
     */
    fun delete(entity: Entity): Long {
        // try to delete the object
        val response = client.delete(
            this.getEndpoint(),
            parser.toJson(entity),
            "application/json; charset=utf-8"
        )

        // in case of error, notify it
        if (!response.isSuccessful)
            throw HttpException(response.code)

        // return the id of the object
        return response.body.string().toLong()
    }
}
